@@ -5,6 +5,7 @@
 
 #define _GNU_SOURCE
 #include <argp.h>
+#include <limits.h>
 #include <stdlib.h>
 #include <string.h>
 #include <syslog.h>
@@ -60,7 +61,7 @@ const struct argp_option options[] = {
      "set environment variables per ARG"},
    { "uid",           'u', "UID",  0, "run as UID within container" },
    { "unset-env",      -7, "GLOB", 0, "unset environment variable(s)" },
-   { "verbose",       'v', 0,      0, "be more verbose (debug if repeated)" },
+   { "verbose",       'v', 0,      0, "be more verbose (can be repeated)" },
    { "version",       'V', 0,      0, "print version and exit" },
    { "write",         'w', 0,      0, "mount image read-write"},
    { 0 }
@@ -109,7 +110,7 @@ int main(int argc, char *argv[])
           argv_to_string(argv));
 #endif
 
-   verbose = 1;  // in charliecloud.h
+   verbose = LL_INFO;  // in ch_misc.c
    args = (struct args){
       .c = (struct container){ .binds = list_new(sizeof(struct bind), 0),
                                .ch_ssh = false,
@@ -180,14 +181,14 @@ int main(int argc, char *argv[])
    for (int i = 0; i < argc - arg_next; i++)
       c_argv[i] = argv[i + arg_next];
 
-   INFO("verbosity: %d", verbose);
-   INFO("image: %s", args.c.img_path);
-   INFO("newroot: %s", args.c.newroot);
-   INFO("container uid: %u", args.c.container_uid);
-   INFO("container gid: %u", args.c.container_gid);
-   INFO("join: %d %d %s %d", args.c.join, args.c.join_ct, args.c.join_tag,
-        args.c.join_pid);
-   INFO("private /tmp: %d", args.c.private_tmp);
+   VERBOSE("verbosity: %d", verbose);
+   VERBOSE("image: %s", args.c.img_path);
+   VERBOSE("newroot: %s", args.c.newroot);
+   VERBOSE("container uid: %u", args.c.container_uid);
+   VERBOSE("container gid: %u", args.c.container_gid);
+   VERBOSE("join: %d %d %s %d", args.c.join, args.c.join_ct, args.c.join_tag,
+           args.c.join_pid);
+   VERBOSE("private /tmp: %d", args.c.private_tmp);
 
    containerize(&args.c);
    fix_environment(&args);
@@ -216,7 +217,7 @@ void fix_environment(struct args *args)
               && !strstr(old_value, ":/bin")) {
       T_ (1 <= asprintf(&new_value, "%s:/bin", old_value));
       Z_ (setenv("PATH", new_value, 1));
-      INFO("new $PATH: %s", new_value);
+      VERBOSE("new $PATH: %s", new_value);
    }
 
    // $TMPDIR: Unset.
@@ -270,13 +271,13 @@ int join_ct(int cli_ct)
    char *ev_name, *ev_value;
 
    if (cli_ct != 0) {
-      INFO("join: peer group size from command line");
+      VERBOSE("join: peer group size from command line");
       j = cli_ct;
       goto end;
    }
 
    if (get_first_env(JOIN_CT_ENV, &ev_name, &ev_value)) {
-      INFO("join: peer group size from %s", ev_name);
+      VERBOSE("join: peer group size from %s", ev_name);
       j = parse_int(ev_value, true, ev_name);
       goto end;
    }
@@ -294,18 +295,18 @@ char *join_tag(char *cli_tag)
    char *ev_name, *ev_value;
 
    if (cli_tag != NULL) {
-      INFO("join: peer group tag from command line");
+      VERBOSE("join: peer group tag from command line");
       tag = cli_tag;
       goto end;
    }
 
    if (get_first_env(JOIN_TAG_ENV, &ev_name, &ev_value)) {
-      INFO("join: peer group tag from %s", ev_name);
+      VERBOSE("join: peer group tag from %s", ev_name);
       tag = ev_value;
       goto end;
    }
 
-   INFO("join: peer group tag from getppid(2)");
+   VERBOSE("join: peer group tag from getppid(2)");
    T_ (1 <= asprintf(&tag, "%d", getppid()));
 
 end:
@@ -323,11 +324,12 @@ int parse_int(char *s, bool extra_ok, char *error_tag)
 
    errno = 0;
    l = strtol(s, &end, 10);
-   Tf (errno == 0, error_tag);
    Ze (end == s, "%s: no digits found", error_tag);
+   Ze (errno == ERANGE || l < INT_MIN || l > INT_MAX,
+       "%s: out of range", error_tag);
+   Tf (errno == 0, error_tag);
    if (!extra_ok)
       Te (*end == 0, "%s: extra characters after digits", error_tag);
-   Te (l >= INT_MIN && l <= INT_MAX, "%s: out of range", error_tag);
    return (int)l;
 }
 
@@ -432,7 +434,7 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
       break;
    case 'v':
       verbose++;
-      Te(verbose <= 4, "--verbose can be specified at most thrice");
+      Te(verbose <= 3, "--verbose can be specified at most thrice");
       break;
    case 'w':
       args->c.writable = true;
