@@ -15,8 +15,12 @@
 /* Log the current UIDs. */
 #define LOG_IDS log_ids(__func__, __LINE__)
 
-/* Test some value, and if it's not what we expect, exit with an error. These
-   are macros so we have access to the file and line number.
+/* C99 does not have noreturn or _Noreturn (those are C11), but GCC, Clang,
+   and hopefully others support the following extension. */
+#define noreturn __attribute__ ((noreturn))
+
+/* Test some value, and if it's not what we expect, exit with a fatal error.
+   These are macros so we have access to the file and line number.
 
                      verify x is true (non-zero); otherwise print then exit:
      T_ (x)            default error message including file, line, errno
@@ -50,18 +54,19 @@
    FIXME: It would be nice if we could collapse these to fewer macros.
    However, when looking into that I ended up in preprocessor black magic
    (e.g. https://stackoverflow.com/a/2308651) that I didn't understand. */
-#define T_(x)      if (!(x)) msg(0, __FILE__, __LINE__, errno, NULL)
-#define Tf(x, ...) if (!(x)) msg(0, __FILE__, __LINE__, errno, __VA_ARGS__)
-#define Te(x, ...) if (!(x)) msg(0, __FILE__, __LINE__, 0, __VA_ARGS__)
-#define Z_(x)      if (x)    msg(0, __FILE__, __LINE__, errno, NULL)
-#define Zf(x, ...) if (x)    msg(0, __FILE__, __LINE__, errno, __VA_ARGS__)
-#define Ze(x, ...) if (x)    msg(0, __FILE__, __LINE__, 0, __VA_ARGS__)
+#define T_(x)      if (!(x)) msg_fatal(__FILE__, __LINE__, errno, NULL)
+#define Tf(x, ...) if (!(x)) msg_fatal(__FILE__, __LINE__, errno, __VA_ARGS__)
+#define Te(x, ...) if (!(x)) msg_fatal(__FILE__, __LINE__, 0, __VA_ARGS__)
+#define Z_(x)      if (x)    msg_fatal(__FILE__, __LINE__, errno, NULL)
+#define Zf(x, ...) if (x)    msg_fatal(__FILE__, __LINE__, errno, __VA_ARGS__)
+#define Ze(x, ...) if (x)    msg_fatal(__FILE__, __LINE__, 0, __VA_ARGS__)
 
-#define FATAL(...)   msg(0, __FILE__, __LINE__, 0, __VA_ARGS__);
-#define WARNING(...) msg(1, __FILE__, __LINE__, 0, __VA_ARGS__);
-#define INFO(...)    msg(2, __FILE__, __LINE__, 0, __VA_ARGS__);
-#define VERBOSE(...) msg(3, __FILE__, __LINE__, 0, __VA_ARGS__);
-#define DEBUG(...)   msg(4, __FILE__, __LINE__, 0, __VA_ARGS__);
+#define FATAL(...)   msg_fatal(      __FILE__, __LINE__, 0, __VA_ARGS__);
+#define WARNING(...) msg(LL_WARNING, __FILE__, __LINE__, 0, __VA_ARGS__);
+#define INFO(...)    msg(LL_INFO,    __FILE__, __LINE__, 0, __VA_ARGS__);
+#define VERBOSE(...) msg(LL_VERBOSE, __FILE__, __LINE__, 0, __VA_ARGS__);
+#define DEBUG(...)   msg(LL_DEBUG,   __FILE__, __LINE__, 0, __VA_ARGS__);
+#define TRACE(...)   msg(LL_TRACE,   __FILE__, __LINE__, 0, __VA_ARGS__);
 
 
 /** Types **/
@@ -84,10 +89,17 @@ struct env_delta {
    } arg;
 };
 
+enum log_level { LL_FATAL =   -2,  // minimum number of -v to print the msg
+                 LL_WARNING = -1,
+                 LL_INFO =     0,
+                 LL_VERBOSE =  1,
+                 LL_DEBUG =    2,
+                 LL_TRACE =    3 };
+
 
 /** External variables **/
 
-extern int verbose;
+extern enum log_level verbose;
 extern char *host_tmp;
 extern char *username;
 
@@ -105,8 +117,10 @@ void list_append(void **ar, void *new, size_t size);
 void *list_new(size_t size, size_t ct);
 void log_ids(const char *func, int line);
 void mkdirs(const char *base, const char *path, char **denylist);
-void msg(int level, const char *file, int line, int errno_,
+void msg(enum log_level level, const char *file, int line, int errno_,
          const char *fmt, ...);
+noreturn void msg_fatal(const char *file, int line, int errno_,
+                        const char *fmt, ...);
 bool path_exists(const char *path, struct stat *statbuf, bool follow_symlink);
 unsigned long path_mount_flags(const char *path);
 void path_split(const char *path, char **dir, char **base);
