@@ -220,8 +220,6 @@ class Main_Loop(lark.Visitor):
             else:
                ch.FATAL("first instruction must be ARG or FROM")
          inst.execute()
-         if (image_i != -1):
-            images[image_i].metadata_save()
          self.instruction_ct += inst.execute_increment
 
 
@@ -260,6 +258,13 @@ class Instruction(abc.ABC):
    def execute(self):
       if (not cli.dry_run):
          self.execute_()
+      if (image_i != -1):
+         # We have no use for the history fields other than upload them back
+         # to the registry at push time, so leave it in the OCI format.
+         images[image_i].metadata["history"].append(
+            { "created": ch.now_utc_iso8601(),
+              "created_by": "%s %s" % (self.str_name(), self.str_()) })
+         images[image_i].metadata_save()
 
    @abc.abstractmethod
    def execute_(self):
@@ -689,17 +694,18 @@ class Run(Instruction):
       exit_code = ch.ch_run_modify(rootfs, cmd, env.env_build, env.workdir,
                                    cli.bind, fail_ok=True)
       if (exit_code != 0):
+         msg = "build failed: RUN command exited with %d" % exit_code
          if (cli.force):
             if (isinstance(fakeroot_config, fakeroot.Fakeroot_Noop)):
-               ch.ERROR("build failed: --force specified, but no suitable config found")
+               ch.FATAL(msg, "--force specified, but no suitable config found")
             else:
-               pass  # we did init --force OK but the build still failed
+               ch.FATAL(msg)  # --force inited OK but the build still failed
          elif (not cli.no_force_detect):
             if (fakeroot_config.init_done):
-               ch.ERROR("build failed: --force may fix it")
+               ch.FATAL(msg, "--force may fix it")
             else:
-               ch.ERROR("build failed: current version of --force wouldn't help")
-         ch.FATAL("build failed: RUN command exited with %d" % exit_code)
+               ch.FATAL(msg, "current version of --force wouldn't help")
+         assert False, "unreachable code reached"
 
    def str_(self):
       return str(self.cmd)
